/*
 * Copyright (c) 2020 - 2023 Andri Yngvason
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#pragma once

#include "stream.h"

#include <aml.h>

static inline void stream__poll_r(struct stream* self)
{
	aml_set_event_mask(self->handler, AML_EVENT_READ);
}

static inline void stream__poll_w(struct stream* self)
{
	aml_set_event_mask(self->handler, AML_EVENT_WRITE);
}

static inline void stream__poll_rw(struct stream* self)
{
	aml_set_event_mask(self->handler, AML_EVENT_READ | AML_EVENT_WRITE);
}

void stream_req__finish(struct stream_req* req, enum stream_req_status status);
void stream__remote_closed(struct stream* self);
